/*
 * Name: pluto.c
 *
 * Description:
 *
 * This driver will be used in combination with a fuzzer and code coverage tool
 * to illustrate how to find missed code blocks and how untrusted input data can
 * be purmutated to hit them. 
 *
 */

#include <ntddk.h>

NTSTATUS Create( PDEVICE_OBJECT, PIRP);
NTSTATUS Read( PDEVICE_OBJECT, PIRP);
NTSTATUS Write( PDEVICE_OBJECT, PIRP) ; 
NTSTATUS Shutdown( PDEVICE_OBJECT, PIRP );
NTSTATUS IoCtl( PDEVICE_OBJECT, PIRP );
VOID Unload( PDRIVER_OBJECT );

#define BUFFERSIZE 512

typedef struct _driver_extension
{
	CHAR buffer[BUFFERSIZE];
} OUR_DRIVER_EXTENSION, *POUR_DRIVER_EXTENSION;

typedef struct _packet
{
	ULONG type;
	ULONG len;
	union
	{
		struct db
		{
			ULONG Dest;
			ULONG Source;
		};
		ULONG Code;
	};
} PACKET, *PPACKET;

OUR_DRIVER_EXTENSION* driverExtension = NULL;

//
// This is where the driver begins
//
NTSTATUS DriverEntry( PDRIVER_OBJECT DriverObject, PUNICODE_STRING RegistryPath ) 
{
	UNICODE_STRING nameString, linkString;
	PDEVICE_OBJECT deviceObject;
	NTSTATUS status;

	RtlInitUnicodeString(&nameString, L"\\Device\\Uninformed");

	status = IoCreateDevice(
		DriverObject, 
		sizeof(65533),
		&nameString, 
		0, 
		0, 
		FALSE, 
		&deviceObject);

	if (!NT_SUCCESS(status))
		return status;

	deviceObject->Flags |= DO_DIRECT_IO;
	deviceObject->Flags &= ~DO_DEVICE_INITIALIZING;

	RtlInitUnicodeString(&linkString, L"\\DosDevices\\Uninformed");

	status = IoCreateSymbolicLink (&linkString, &nameString);

	if (!NT_SUCCESS(status)) 
	{
		IoDeleteDevice (DriverObject->DeviceObject);
		return status;
	}

	DriverObject->MajorFunction[IRP_MJ_CREATE]		= Create;
	DriverObject->MajorFunction[IRP_MJ_READ]		= Read;
	DriverObject->MajorFunction[IRP_MJ_WRITE]		= Write;
	DriverObject->MajorFunction[IRP_MJ_SHUTDOWN]		= Shutdown;
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]	= IoCtl;
	DriverObject->DriverUnload = Unload;

	driverExtension = ExAllocatePool(NonPagedPool, sizeof ( OUR_DRIVER_EXTENSION));

	if(!driverExtension) 
		return STATUS_INSUFFICIENT_RESOURCES;

	RtlZeroMemory( driverExtension->buffer, BUFFERSIZE);

	return STATUS_SUCCESS;
}

// 
// This is called when our driver is unloaded
//
VOID Unload( PDRIVER_OBJECT DriverObject ) 
{
	UNICODE_STRING linkString;

	ExFreePool(driverExtension);
	RtlInitUnicodeString (&linkString, L"\\DosDevices\\Uninformed");
	IoDeleteSymbolicLink (&linkString);
	IoDeleteDevice(DriverObject->DeviceObject);

	return;
}

//
// This is called when the device is opened
//
NTSTATUS Create (PDEVICE_OBJECT DeviceObject, IN PIRP Irp) 
{
	IoCompleteRequest(Irp,IO_NO_INCREMENT);

	return	STATUS_SUCCESS;
}

// 
// This is called when the device is read from
//
NTSTATUS Read(PDEVICE_OBJECT DeviceObject, IN PIRP Irp) 
{
	PUCHAR				              currentAddress;
	PIO_STACK_LOCATION				  irpStack;

	if (!driverExtension) 
	{
		IoCompleteRequest(Irp, IO_NO_INCREMENT);

		return STATUS_INSUFFICIENT_RESOURCES;
	}

	irpStack = IoGetCurrentIrpStackLocation(Irp);

	if (irpStack->MajorFunction == IRP_MJ_READ) 
	{
		currentAddress = MmGetSystemAddressForMdlSafe(Irp->MdlAddress, NormalPagePriority);

		if (!currentAddress) 
		{
			IoCompleteRequest(Irp, IO_NO_INCREMENT);

			return STATUS_SUCCESS;
		}

		RtlMoveMemory(currentAddress, 
		driverExtension->buffer+irpStack->Parameters.Read.ByteOffset.LowPart,
		irpStack->Parameters.Read.Length);
	}

	IoCompleteRequest(Irp, IO_NO_INCREMENT);

	return STATUS_SUCCESS;
}

//
// This is called when the device is written to
//
NTSTATUS Write( PDEVICE_OBJECT DeviceObject, PIRP Irp ) 
{
	PUCHAR currentAddress;
	PIO_STACK_LOCATION irpStack;
	PPACKET Packet = NULL;
	ULONG Dest, Source;

	if (!driverExtension) 
	{
		IoCompleteRequest(Irp, IO_NO_INCREMENT);
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	irpStack = IoGetCurrentIrpStackLocation(Irp);

	if (irpStack->MajorFunction == IRP_MJ_WRITE) 
	{
		currentAddress = MmGetSystemAddressForMdlSafe(Irp->MdlAddress, NormalPagePriority);

		if (!currentAddress) 
		{
			IoCompleteRequest(Irp, IO_NO_INCREMENT);
			return STATUS_SUCCESS;
		}

		RtlMoveMemory(driverExtension->buffer,
				currentAddress, irpStack->Parameters.Write.Length);
		
		Packet = (PPACKET)driverExtension->buffer;
		
#ifdef DEBUG
		DbgPrint("Packet->type = %x\n", Packet->type);
		DbgPrint("Packet->len = %x\n", Packet->len);
#endif
	
		switch(Packet->type)
		{
			case 0xdeadbeef:
				goto bad;
			default:
			{
				// 
				// Unsupported types get ignored 
				//
				break;
			}
		}
	}

	IoCompleteRequest(Irp, IO_NO_INCREMENT);

	return STATUS_SUCCESS;
bad:
	//
	// Write a specified value at a specified address 
	//
	Dest = Packet->Dest;
	Source = Packet->Source;

	RtlMoveMemory( (PVOID)Dest, (PVOID)Source, 4 );

	IoCompleteRequest(Irp, IO_NO_INCREMENT);

	return STATUS_SUCCESS;
}

// 
// This get called when we are passed an IOCTL
//
NTSTATUS IoCtl(PDEVICE_OBJECT DeviceObject, IN PIRP Irp) 
{
	IoCompleteRequest(Irp, IO_NO_INCREMENT);

	return STATUS_SUCCESS;
}

// 
// This gets called when the device is shutdown
//
NTSTATUS Shutdown(PDEVICE_OBJECT DeviceObject, IN PIRP Irp) 
{
	IoCompleteRequest(Irp, IO_NO_INCREMENT);

	return STATUS_SUCCESS;
}

